/*
 * userland + libusb specific functions
 * $Id: ifp_os_libusb.c,v 1.2 2005/08/25 04:10:54 jim-campbell Exp $
 *
 * Copyright (C) 2004 Geoff Oakham <oakhamg@users.sourceforge.net>
 */

#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <libusb-1.0/libusb.h>
#include <iconv.h>
#include <locale.h>
#include <langinfo.h>

#include "ifp.h"
#include "ifp_os.h"
#include "prim.h"

static int local_iconv(char const * i_type, char const * o_type,
                       char * ob, int max_o, const char * ib, int max_i)
{
    int i=0, e=0;
    char * ibb = (char *)ib;
    char * obb = (char *)ob;
    size_t i_n = max_i;
    size_t o_n = max_o;
    iconv_t ICONV;
    size_t r;

    if (i_type == NULL) {
        ifp_err("itype is NULL");
        i = -EINVAL;
        goto err;
    }

    if (o_type == NULL) {
        ifp_err("otype is NULL");
        i = -EINVAL;
        goto err;
    }

    ICONV = iconv_open(o_type, i_type);
    if (ICONV == (iconv_t)-1) {
        i = -errno;
        if(i==-EINVAL) {
            ifp_err_i(i, "conversion not supported by system");
        } else {
            ifp_err_i(i, "couldn't open conversion handle");
        }
        goto err;
    }

    r = iconv(ICONV, &ibb, &i_n, &obb, &o_n);
    if (r == (size_t)-1) {
        i = -errno;
        ifp_err_i(i, "problem converting, i_n is %ld, o_n is %ld, r = %ld", i_n, o_n, r);
        goto err2;
    }

    err2:
    e = iconv_close(ICONV);
    if (e) {
        e = -errno;
        ifp_err_i(e, "couldn't close conversion");
        i = i?i:e;
    }
    err:
    return i;
}

int ifp_locale_to_utf16(char * ob, int max_o, const char * ib, int max_i)
{
    int i;

    i = local_iconv(nl_langinfo(CODESET), "UTF-16LE", ob, max_o, ib, max_i);
    ifp_err_jump(i, err, "conversion failed");

    err:
    return i;
}

static int utf16_sizeof(uint16_t * data, int max) {
    int i=0, j;
    for (j=0; j!=max && i==0; j++) {
        if (data[j] == 0) {
            return j;
        }
    }
    return max;
}

int ifp_utf16_to_locale(char * ob, int max_o, const char * ib, int max_i)
{
    int i;
    int n;

    n = utf16_sizeof((uint16_t *)ib, max_i/sizeof(uint16_t));

    i = local_iconv("UTF-16LE", nl_langinfo(CODESET), ob, max_o, ib, (n+1)*sizeof(uint16_t));
    //i = local_iconv("UTF-16LE", nl_langinfo(CODESET), ob, max_o, ib, max_i);
    ifp_err_jump(i, err, "conversion failed");

    err:
    return i;
}

int ifp_os_sleep(int ms) {
    struct timespec ts;

    ts.tv_sec = ms / 1000;
    ts.tv_nsec = (ms % 1000)*1000000;
    return nanosleep(&ts, NULL);
}

#define MAX_TRIES 8

int ifp_os_control_send(struct ifp_device * dev, int command, int arg1, int arg2,
                        int * r1, int * r2)
{
    struct libusb_device_handle * usbdev = dev->device;
    uint8_t ctl[8];
    int c_size;
    int i;
    int max_try = MAX_TRIES; IFP_BUG_ON(r1 == NULL); c_size = r2 ? 8 : 4;

    //usb PIPE/stall error detection and correction.
    //
    // AFAIK, only Linux is affected by this bug, so I think it's important
    // that the code get really noisy if this happens on another platform.
    // (Hence the ifdefs.)
    do {
        i = libusb_control_transfer(usbdev, IFP_REQ_TYPE, command,
                                    arg1, arg2, ctl, c_size, IFP_TIMEOUT);

        max_try--;
        if (i == -EPIPE) {
#ifdef linux
            if (command != IFP_FILE_DOWNLOAD || max_try != (MAX_TRIES - 1)) {
#endif
                ifp_err("found an EPIPE error not previously documented. "
                        "command=%02x try=%d", command, MAX_TRIES - max_try);
#ifdef linux
            }

            if (command == IFP_FILE_DOWNLOAD) {
                dev->download_pipe_errors++;
            }
#endif
            ifp_os_sleep(100); //I'm still not convienced this helps anything
        }
    } while(i == -EPIPE && max_try > 0);
    if (i < 0) {
        ifp_err_i(i, "error %s ifp control code the command %0x (%d, %d)."
                  " ctl[%d]  Returned %d.",
                  "sending", command, arg1, arg2, c_size, i);
        return i;
    } else if (i != c_size) {
        if (command == IFP_FILE_DOWNLOAD) { dev->alt_readcount++; }
        ifp_err_i(i, "warning: unexpected error value.. I expected %d.",
                  c_size);
    } else {
        if (command == IFP_FILE_DOWNLOAD) { dev->alt_readcount++; }
        i = 0;
    }

    if (r2) {
        *r2 = ifp_os_le32_to_cpup(ctl + 4);
    }
    *r1 = ifp_os_le32_to_cpup(ctl);

    return i;
}

int ifp_os_push(struct ifp_device * dev, void * p, int n)
{
    int i = 0;
    int actual_length;

    if (dev->bulk_to & 0x80) {
        /* libusb-0.1 on BSD strangely fix up a write request to endpoint
         * 0x81 to be to endpoint 0x01. do the same thing here, but
         * warn about this silly behaviour. */
        ifp_wrn("endpoint %x has excessive IN direction bit, fixing", dev->bulk_to);
        dev->bulk_to &= ~0x80;
    }

    i = libusb_bulk_transfer(dev->device, dev->bulk_to, p, n,
                             &actual_length, IFP_TIMEOUT);
    if (i == 0 || (i == LIBUSB_ERROR_TIMEOUT && actual_length > 0))
        i = actual_length;

    if (i < 0) {
        ifp_err_i(i, "usb_bulk_msg failed");
        return i;
    } else if (i != n) {
        ifp_wrn(" usb_bulk_msg wrote %d bytes instead of %d.", i, n);
    }

    return 0;
}

int ifp_os_pop(struct ifp_device * dev, void * p, int n)
{
    int i = 0;
    int actual_length;
    memset(p, 0, n);

    if (!(dev->bulk_from & 0x80)) {
        /* libusb-0.1 will strangely fix up a read request from endpoint
             * 0x01 to be from endpoint 0x81. do the same thing here, but
             * warn about this silly behaviour. */
        ifp_wrn("endpoint %x is missing IN direction bit, fixing", dev->bulk_from);
        dev->bulk_from |= 0x80;
    }

    //ifp_wrn("endpoint %x size %d timeout %d", dev->bulk_from, n, IFP_TIMEOUT);
    i = libusb_bulk_transfer(dev->device, dev->bulk_from, p, n,
                             &actual_length, IFP_TIMEOUT);

    /* if we timed out but did transfer some data, report as successful short
         * read. FIXME: is this how libusb-0.1 works? */
    if (i == 0 || (i == LIBUSB_ERROR_TIMEOUT && actual_length > 0))
        i =actual_length;

    if (i < 0) {
        ifp_err_i(i, "usb_bulk_msg failed");
        return i;
    } else if (i != n) {
        ifp_wrn("read in %d bytes instead of %d",i,n);
    }

    return 0;
}

int ifp_os_init(struct ifp_device * dev, void * device_handle)
{
    int i = 0;
    libusb_device_handle *ldev = device_handle;
    libusb_device *usbdev = libusb_get_device(ldev);
    struct libusb_device_descriptor desc;
    struct libusb_config_descriptor *config;
    const struct libusb_interface_descriptor *idesc;
    int address0, address1;

    //FIXME: move this to a more appropriate place.
    setlocale(LC_ALL, "");

    libusb_get_device_descriptor(usbdev, &desc);

    dev->model = desc.idProduct;
    dev->device = device_handle;

    libusb_get_config_descriptor(usbdev, 0, &config);

    /* sanity checks that we're looking at an acceptable device */
    if (/*dev->descriptor.bNumConfigurations != 1 || TODO: ?*/
            config->bNumInterfaces != 1 ||
            config->interface[0].num_altsetting != 1)
        return 1;

    idesc = &config->interface[0].altsetting[0];

    if (idesc->bNumEndpoints != 2)
        return 1;

    /* grab the pointers */
    //const static int ENDPOINT_BULK_IN=0x82; /* endpoint 0x81 address for IN */
    //const static int ENDPOINT_BULK_OUT=0x02; /* endpoint 1 address for OUT */

    address0 = (idesc->endpoint[0]).bEndpointAddress;
    address1 = (idesc->endpoint[1]).bEndpointAddress;

    if (address0 & 0x0080) {
        IFP_BUG_ON(address1 & 0x0080);
        dev->bulk_from = address0;
        dev->bulk_to = address1;
    } else {
        IFP_BUG_ON((address1 & 0x0080)==0);
        dev->bulk_from = address1;
        dev->bulk_to = address0;
    }

    return i;
}
int ifp_os_finalize(struct ifp_device * dev)
{

    //release happens somewhere else
    dev->device = NULL;
    //libusb_exit(NULL);
    return 0;
}

void * ifp_find_device(void)
{
    const int iRiver_Vendor = 0x4102;
    int listSize;
    int cnt;
    int ret;
    libusb_device **list;
    libusb_device *found = NULL;
    libusb_device_handle *handle = NULL;

    ret = libusb_init(NULL);
    if (ret != 0)
        return NULL;

    listSize = libusb_get_device_list(NULL, &list);

    if (listSize < 0)
        return NULL;

    for (cnt = 0; cnt < listSize; cnt++)
    {
        libusb_device *dev = list[cnt];
        struct libusb_device_descriptor desc;

        ret = libusb_get_device_descriptor(dev, &desc);
        if (ret < 0) {
            fprintf(stderr, "failed to get device descriptor");
            return NULL;
        }

        if (desc.idVendor  == iRiver_Vendor)
        {
            if ((desc.idProduct & 0xFF00) == 0x1100)
            {
                ifp_info("ignoring device with UMS firmware.");
            }
            else if ((desc.idProduct & 0xFF00) != 0x1100)
            {
                found = dev;
                break;
            }
        }
    }

    if (found) {
        ret = libusb_open(found, &handle);
        if (ret != 0) {
            handle = NULL;
        }
    }

    libusb_free_device_list(list, 1);

    return handle;
}


int ifp_release_device(void * dh)
{
    if (dh != NULL)
    {
        libusb_close(dh);
        //free((libusb_device_handle *)dh);
    }
    return 0;
}



